"""Custom types."""
from pathlib import Path
from typing import TypedDict, List, Any


class ArgumentInfo(TypedDict):
    """Info about an argument of a python function."""
    name: str
    type: str
    default: Any


class FunctionInfo(TypedDict):
    """Info about a python function."""
    name: str
    arguments: List[ArgumentInfo]
    return_type: str
    docstring: str
    is_private: bool


class FileInfo(TypedDict):
    """Info about a .py file."""
    name: str
    path: Path
    module: str
    package: str
    test_name: str
    test_path: Path
    functions: List[FunctionInfo]



"""Main module which contains all necessary functions to create test stubs."""
import os
from pathlib import Path
from typing import List

import typer
from jinja2 import Environment, FileSystemLoader
from redbaron import RedBaron

from pytest_codegen.types import FileInfo, FunctionInfo, ArgumentInfo

app = typer.Typer()
env = None


@app.command()
def main(package: str, test_dir: str = "./tests/unittests") -> None:
    """Run the program."""
    package_path = Path(package).resolve()
    test_dir_path = Path(test_dir).resolve()

    files = _collect_files(package_path)
    file_infos = _collect_information(files, package_path, test_dir_path)
    _create_directories(file_infos)
    _create_files(file_infos)
    _write_files(file_infos)


def _write_files(files: List[FileInfo]) -> None:
    """Write the test stubs to the test file."""
    function_template = _get_environment().get_template("function.template")
    import_template = _get_environment().get_template("import.template")
    for file in files:

        with open(file["test_path"] / file["test_name"], "r") as f:
            root = RedBaron(f.read() or '"""Test module."""\n')

        existing_functions = [function.name for function in root.find_all("def")]

        for function in file["functions"]:
            if f"test_{function['name']}" not in existing_functions:
                root.extend(["\n", "\n"])
                root.append(function_template.render(**function))

        root.insert(1, import_template.render(**file))
        root.insert(1, "import pytest")

        with open(file["test_path"] / file["test_name"], "w") as f:
            f.write(root.dumps())


def _create_files(files: List[FileInfo]) -> None:
    """Create the test files themselves."""
    for file in files:
        try:
            with open(file["test_path"] / file["test_name"], "x"):
                pass
        except FileExistsError:
            pass


def _create_directories(files: List[FileInfo]) -> None:
    """Create directories for test files."""
    for file in files:
        if not os.path.exists(file["test_path"]):
            os.makedirs(file["test_path"], exist_ok=True)


def _get_environment() -> Environment:
    """Get the jinja2 template environment."""
    global env
    if not env:
        loader = FileSystemLoader(searchpath=Path(__file__).parent / Path("templates"))
        env = Environment(loader=loader)
    return env


def _collect_information(
    files: List[Path], base_path: Path, test_dir_path: Path
) -> List[FileInfo]:
    """Collect information about all files in the list."""
    file_info_list = []
    for file in files:
        file_info = FileInfo(
            name=file.name,
            module=file.name.split(".")[0],
            package=".".join(
                os.path.relpath(file, base_path.parent)[:-3].split(os.sep)
            ),
            path=file.parent,
            test_path=test_dir_path / Path(os.path.relpath(file, base_path)).parent,
            functions=[],
        )  # type: ignore

        file_info["test_name"] = (
            _get_environment().get_template("filename.template").render(**file_info)
        )

        root_fst = RedBaron(file.read_text())

        # collect function information
        for function in root_fst.find_all("def"):

            # TODO: properly find below fields
            function_info = FunctionInfo(
                name=function.name,
                is_private=function.name.startswith("_"),
                docstring=function[0].find("string"),
                return_type=str(function.return_annotation.value),
                arguments=[],
            )

            #  collect function's argument information
            for argument in function.arguments:
                argument_info = ArgumentInfo(
                    name=str(argument.name),
                    default=argument.value.value
                    if argument.value is not None
                    else None,
                    type=argument.annotation.value
                    if argument.annotation is not None
                    else None,
                )
                function_info["arguments"].append(argument_info)

            file_info["functions"].append(function_info)
        if file_info["functions"]:
            file_info_list.append(file_info)

    return file_info_list


def _collect_files(path: Path) -> List[Path]:
    """Collect all .py files in a package."""
    files_list: List = []
    if os.path.isdir(path):
        files = os.listdir(path)
        for file in files:
            files_list.extend(_collect_files((Path(path) / file).absolute()))
    else:
        if path.name.endswith(".py") and not path.name.startswith("__"):
            files_list.append(Path(path).absolute())

    return files_list



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytest_codegen']

package_data = \
{'': ['*'], 'pytest_codegen': ['templates/*']}

install_requires = \
['jinja2>=2.11.2,<3.0.0', 'redbaron>=0.9.2,<0.10.0', 'typer>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['ptc = pytest_codegen.__main__:entry',
                     'pytest-codegen = pytest_codegen.__main__:entry']}

setup_kwargs = {
    'name': 'pytest-codegen',
    'version': '0.0.3',
    'description': 'Automatically create pytest test signatures',
    'long_description': '# pytest-codegen\nPytest-codgen will statically analyze your code to create pytest function stubs.\n\nBuilt with [redbaron](https://redbaron.readthedocs.io/en/latest/), [typer](https://typer.tiangolo.com/) and [jinja2](https://jinja.palletsprojects.com/en/2.11.x/).\n\n\n## Goal\nFirst working version\n\n### Future Goals\n- Create templates for tests\n- More customization\n\n\n## Disclaimer\nThis tool is currently in pre-alpha/experimental phase. Usable version will be ^0.1.x\n\n\n## Installation\n\n```\npip install pytest-codegen\n```\n\n## Usage\n\nCheck the supported commands with\n```\npytest-codegen --help\n```\nor if you are lazy like me\n```\nptc --help\n```\n\n## Suggestions & Contribution\n\nEvery suggestion and contribution is welcome!\n\nSet everything up with [poetry](https://python-poetry.org/) and you are good to go :)\n```\npip install poetry\n```\n```\npoetry install\n```\n\n## Ressources\n- [pytest-codegen on pypi](https://pypi.org/project/pytest-codegen/)\n- [redbaron on pypi](https://pypi.org/project/redbaron/)\n- [typer on pypi](https://pypi.org/project/typer/)\n- [jinja2 on pypi](https://pypi.org/project/Jinja2/)\n\n## License\nThis project is licensed under the terms of the MIT license.',
    'author': 'Jeremy Schiemann',
    'author_email': 'jeremy.schiemann@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jeremyschiemann/pytest-codegen',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
